// -----------------------------------------------------------------------------------------
//
// umFish20.CPP                                                       Ulrich Mller   2.3.00
//                                                                    Procida 400NT VC++ 6.0
// Basisfunktionen fr das fischertechnik Interface (seriell und parallel)
// Der Interfacestatus wird asynchron ermittelt(E- und A-Eingnge) bzw. gesetzt (M-Ausgnge)
// umOpenInterface startet dazu den MultiMediaTimer mit PollInterface als CallBack-Routine
// Die Statusdaten werden ber den ftiDCB ausgetauscht, die "s"-Funktionen entfallen 
// deswegen (GetInputs, SetMotors, ClearMotors, GetAnalog()
// 
// -----------------------------------------------------------------------------------------

#include <windows.h>
#include <mmsystem.h>
#include <ole2.h>
#include "umFish20.H"

DWORD __stdcall umOpenInterface(ftiDCB &Interface, LPCSTR ComName);
DWORD __stdcall umCloseInterface(ftiDCB &Interface);
BOOL  __stdcall umGetInput(ftiDCB &Interface, DWORD InputNr);
DWORD __stdcall umSetMotor(ftiDCB &Interface, DWORD MotorNr, DWORD Direction);
DWORD __stdcall umSetLamp(ftiDCB &Interface, DWORD LampNr, DWORD OnOff);
DWORD __stdcall umGetStatus();
DWORD __stdcall umGetVersion();

DWORD OpenInterfaceCOM(ftiDCB &Interface, LPCSTR ComName);
DWORD OpenInterfaceLPT(ftiDCB &Interface, LPCSTR LptName);
DWORD GetInputs(ftiDCB &Interface);
DWORD GetInputsCOM1(ftiDCB &Interface);
DWORD GetInputsCOM2(ftiDCB &Interface);
DWORD GetInputsLPT(ftiDCB &Interface);
DWORD GetAnalog(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetAnalogCOM1(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetAnalogCOM2(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetAnalogLPT(ftiDCB &Interface, DWORD AnalogNr);
DWORD GetLPT(ftiDCB &Interface);
VOID  SetLPT(ftiDCB &Interface, WORD Wert);


ftiDCB ftiSave;



// PollInterface ---------------------------------------------------------------------------
//
// Zentrales Pollen eines Interfaces, setzen InputStatus, OutputStatus, Analogs
// Fhren eines Impulszhlers (Counters) fr die E-Eingnge
// Die A-Eingnge werden nur bei AnalogScan > 0 gepollt und dann im Wechsel 0/1
// Beim seriellen Interface wird dann der IntputStatus ber das AnalogLesen ermittelt
//
// -----------------------------------------------------------------------------------------

void CALLBACK PollInterface(UINT wTimerID, UINT msg, DWORD DCB, DWORD dw1, DWORD dw2)
{
	DWORD StatusNeu, StatusAlt, StatusDelta, i;

	ftiDCB *d = (ftiDCB*)DCB;

  StatusAlt = d->InputStatus;
	if(StatusAlt == ftiFehler) return;
	if(d->PortID == 0) {                     // --- COM-Interface
		if(d->AnalogScan == 0) StatusNeu = GetInputs(*d);
		else {
		d->Analogs[d->AnalogScan-1] = GetAnalog(*d, d->AnalogScan-1);
		StatusNeu = d->InputStatus;
		if(d->AnalogScan == 1) d->AnalogScan = 2; else d->AnalogScan = 1;
		}
	}
	else {                                   // --- LPT-Interface
		StatusNeu = GetInputs(*d);
		if(d->AnalogScan > 0) {
		d->Analogs[d->AnalogScan-1] = GetAnalog(*d, d->AnalogScan-1);
		if(d->AnalogScan == 1) d->AnalogScan = 2; else d->AnalogScan = 1;
		}
	}
	StatusDelta = StatusAlt ^ StatusNeu;
	d->InputStatus = StatusNeu;
  for (i=1; i <= d->ECount; i++) 
		if((StatusDelta & EMaske[i]) > 0)  d->Counters[i-1]++;
}



// umGetVersion ----------------------------------------------------------------------------
//
DWORD __stdcall umGetVersion(){return(201);}




// umOpenInterface ------------------------------------------------------------
//
// Setzen Ausgangswerte, Open Schnittstelle
//  PortName (ComName)  COM1 - 9 : alle
//                      LPT1 - 3 (Win9x, wenn Standard Port, CreateFile)
//                      LPT (Win9x) PortID zur Identification, kein CreateFile
//                          WinNT in Registry eingestellter Port, 
//                          WinRT.SYS mu laufen (Allow Conflicts reicht)
//
// --------------------------------------------------------------------------

DWORD __stdcall umOpenInterface(ftiDCB &Interface, LPCSTR ComName)
{
	DWORD res;

	res = ftiFehler;
	Interface.hCom = 0;
	Interface.OutputStatus = 0;
	if(ComName == 0) return(ftiFehler);

  if(ComName[0] == 'C' || ComName[0] == 'c')
    res = OpenInterfaceCOM(Interface, ComName);
  else
    res = OpenInterfaceLPT(Interface, ComName);
	if(res == ftiFehler) {ftiSave = Interface; return(ftiFehler);}

  if(Interface.AnalogScan != 1) Interface.AnalogScan = 0;
	if(Interface.ECount != 16) Interface.ECount = 8;
	if(Interface.PortID == 0) {                    // --- serielles Interface
		if(Interface.PollInterval < 10) Interface.PollInterval = 10;
		if((Interface.ECount == 16) && (Interface.PollInterval < 12))
			Interface.PollInterval = 12;
		if((Interface.AnalogScan > 0) && (Interface.PollInterval < 16))
			Interface.PollInterval = 16;	}
	else {                                          // --- paralleles Interface 
		if(Interface.PollInterval < 1) Interface.PollInterval = 1;
		if((Interface.AnalogScan > 0) && (Interface.PollInterval < 20))
			Interface.PollInterval = 100; }


// ------- Start MultiMediaTimer -------------------------------------------------------

	Interface.FID = timeSetEvent(Interface.PollInterval,
		                            0,                     // --- Auflsung in mSek (0 = max)
		                            PollInterface,         // --- CallBack-Routine
                                DWORD(&Interface),     // --- bergabe ftiDCB 
									              TIME_PERIODIC); 
	ftiSave = Interface;
	return(res);
}

// OpenInterfaceCOM ------------------------------------------------------------
// 
// Setzen Ausgangswerte, Open Schnittstelle
//
// --------------------------------------------------------------------------

DWORD OpenInterfaceCOM(ftiDCB &Interface, LPCSTR ComName)
{
  DCB dcb;
  COMMTIMEOUTS cTime;
  DWORD res;
  HANDLE hCom;

// --- ffnen des ComPorts ---

  hCom = CreateFile(ComName,
    GENERIC_READ | GENERIC_WRITE,
    0,                               // --- Open exklusiv
    NULL,                            // --- keine Security-Attribute
    OPEN_EXISTING,                   // --- Open Modus
    0,                               // --- kein overlapped I/O
    NULL);                           // --- kein Template
  
  if (hCom == INVALID_HANDLE_VALUE) return(ftiFehler);

  // --- Einstellen der COM-Parameter ---

    if (!GetCommState(hCom, &dcb)) return(ftiFehler);

    dcb.BaudRate = 9600;
    dcb.ByteSize = 8;
    dcb.Parity   = NOPARITY;
    dcb.StopBits = ONESTOPBIT;

    if (!SetCommState(hCom, &dcb)) return(ftiFehler);

    cTime.ReadIntervalTimeout         = 5;
    cTime.ReadTotalTimeoutMultiplier  = 4;
    cTime.ReadTotalTimeoutConstant    = 30;
    cTime.WriteTotalTimeoutMultiplier = 1;
    cTime.WriteTotalTimeoutConstant   = 10;

    if (!SetCommTimeouts(hCom, &cTime)) return(ftiFehler);

    Interface.hCom         = hCom;
    Interface.PortID       = 0;
    Interface.LPTDelay     = 1;
    Interface.LPTAnalog    = 1;
    Interface.OutputStatus = 0;
    res = GetInputsCOM1(Interface);
    if (res == ftiFehler) {
      CloseHandle(hCom);
			Interface.hCom = 0;
      return(ftiFehler);
    }

    return(res);
}
 
// OpenInterfaceLPT ------------------------------------------------------------
// 
// Setzen Ausgangswerte, Open LPT-Schnittstelle
// Wenn Version = Wondows NT / 2000 wird ber OpenInterfaceRT der WinRT-Treiber geffnet
// sonst wird bei LPTn direkt auf das Interface zugegriffen
//                     (LPT1 : 0x378, 2 : 0x278, 3 : 0x3BC)
//                LPT  ebenfalls ber OpenInterfaceRT der WinRT-Treiber geffnet
//
// -----------------------------------------------------------------------------

DWORD OpenInterfaceLPT(ftiDCB &Interface, LPCSTR LptName)
{
  
  DWORD res;
  HANDLE hLPT;

  if(GetVersion() <0x80000000)          // --- Windows NT       
	{
    return(OpenInterfaceRT(Interface));
	}
	else
	{
// --- ffnen LPT-Port (Schutzmanahme, fr Zugriff ber Get/SetLPT nicht erforderl.

		if(LptName[3] != 0) {
		  hLPT = CreateFile(LptName,
			  GENERIC_READ | GENERIC_WRITE,
			  0,                                 // --- Open eklusiv
			  NULL,                              // --- keine Security-Attribute
			  OPEN_EXISTING,                     // --- Open-Modus
			  0,                                 // --- kein Overlapped
			  NULL);                             // --- kein Template

		  if(hLPT == INVALID_HANDLE_VALUE) return(ftiFehler);

		  switch (LptName[3]) {
		  case '1': Interface.PortID = 0x378; break; 
		  case '2': Interface.PortID = 0x278; break;
		  case '3': Interface.PortID = 0x3BC; break;
		  default:     Interface.PortID = 0x378;
			}
	  	Interface.hCom = hLPT;
		  if(Interface.LPTDelay == 0) Interface.LPTDelay    = 10;
		  if(Interface.LPTAnalog == 0) Interface.LPTAnalog   = 5;
		  Interface.OutputStatus = 0;
		  res = GetInputs(Interface);
		  if(res == ftiFehler) {
			  CloseHandle(hLPT);
			  Interface.hCom = 0;
			  return(ftiFehler);
			}
      return(res);
		}
		else
		{
		return(OpenInterfaceRT(Interface));
		}
	}
}


// umCloseInterface ------------------------------------------------------------
// 
// Schlieen der Schnittstelle
//
// --------------------------------------------------------------------------

DWORD __stdcall umCloseInterface(ftiDCB &Interface)
{
  DWORD res;
	HANDLE hCom = Interface.hCom;
  if(hCom == 0) return(0);


	Interface.OutputStatus = 0;              // --- Lschen M-Ausgnge
	Sleep(4 * Interface.PollInterval);       // --- dem MM-Timer Zeitlassen
	if(Interface.FID != 0) timeKillEvent(Interface.FID); // dann Killen
	                                          // --- Freigabe der Handles
  if (Interface.PortID == 1)	res = CloseInterfaceRT(Interface); 
	else res = CloseHandle(hCom);

  Interface.hCom   = 0;                    // --- Open-Flag -----------------
	Interface.PortID = 0;
	ftiSave = Interface;                     // --- fr ein Close beim Detach
	return(res);
}


// GetInputs -----------------------------------------------------------------
// 
// Setzen Motorstatus, Lesen Inputstatus
//
// -----------------------------------------------------------------------------

DWORD GetInputs(ftiDCB &Interface)
{
  if(Interface.hCom == 0) return(ftiFehler);
	if(Interface.PortID == 0)
		if (Interface.ECount == 8) return(GetInputsCOM1(Interface));
    else return(GetInputsCOM2(Interface));
  else
	  if(Interface.PortID == 1) return(GetInputsRT(Interface)); 
	  else return(GetInputsLPT(Interface));
}

// GetInputsCOM1 -------------------------------------------------------------
//
// Setzen Motorstatus, Lesen Inputstatus
//
// --------------------------------------------------------------------------

DWORD GetInputsCOM1(ftiDCB &Interface)
{
  HANDLE hCom = Interface.hCom;
  unsigned char Befehl[2];
	unsigned char IStatus[1];
  BOOL  res;
  DWORD Geschrieben;
  DWORD Gelesen;

  Befehl[0] = '\xC1';
  Befehl[1] = unsigned char(Interface.OutputStatus);

  res = WriteFile(hCom, Befehl, 2, &Geschrieben, NULL);
  res = ReadFile(hCom, IStatus, 1, &Gelesen, NULL);
  if (Gelesen == 0) {return(ftiFehler);}

  Interface.InputStatus = unsigned char(IStatus[0]);
  return(Interface.InputStatus);
}


// GetInputsCOM2 -------------------------------------------------------------
//
// Setzen Motorstatus, Lesen Inputstatus
//
// --------------------------------------------------------------------------

DWORD GetInputsCOM2(ftiDCB &Interface)
{
  HANDLE hCom = Interface.hCom;
  unsigned char Befehl[3];
	unsigned char IStatus[2];
  BOOL  res;
  DWORD Geschrieben;
  DWORD Gelesen;

  Befehl[0] = '\xC2';
  Befehl[1] = unsigned char(Interface.OutputStatus & 0xFF);
	Befehl[2] = unsigned char((Interface.OutputStatus & 0xFF00)>>8);

  res = WriteFile(hCom, Befehl, 3, &Geschrieben, NULL);
  res = ReadFile(hCom, IStatus, 2, &Gelesen, NULL);
  if (Gelesen == 0) {return(ftiFehler);}

  Interface.InputStatus = unsigned char(IStatus[0]) + (unsigned char(IStatus[1])<<8);
  return(Interface.InputStatus);
}


// GetInputsLPT -------------------------------------------------------------
//
// Setzen Motorstatus, Lesen Inputstatus
//
// --------------------------------------------------------------------------

DWORD GetInputsLPT(ftiDCB &Interface)
{
  DWORD i, IStatus = 0;

// ---- bertragen Motorstatus ---

  for (i=16; i>=1; i--)
  {
    if(Interface.OutputStatus & EMaske[i])
    {                                             // --- Motor luft
      SetLPT(Interface, 0x34);                    // --- DATA-OUT & Trigger
      SetLPT(Interface, 0x3C);                    // --- DATA-OUT, CLOCK, Trigger
    }
    else
    {                                             // --- Motor steht
      SetLPT(Interface, 0x30);                    // --- 0 & Trigger
      SetLPT(Interface, 0x38);                    // --- CLOCK & Trigger
    }
  }
    SetLPT(Interface, 0x39);                      // --- LOAD-OUT, CLOCK, Trigger

// ---- Lesen Inputstatus -------

  SetLPT(Interface, 0x32);                        // --- LOAD-IN
  SetLPT(Interface, 0x3A);                        // --- LOAD & CLOCK
  for (i=8; i>=1; i--)                      // --- Erstes Interface
  {                                               // --- Auslesen einzelner Eingang
    if(GetLPT(Interface) & 0x80) IStatus |= EMaske[i];
    SetLPT(Interface, 0x30);                      // --- null
    SetLPT(Interface, 0x38);                      // --- CLOCK
  }

  for (i=16; i>=9; i--)                           // --- Slave Interface
  {                                               // --- Auslesen einzelner Eingang
    if(GetLPT(Interface) & 0x80) IStatus |= EMaske[i];
    SetLPT(Interface, 0x30);                      // --- null
    SetLPT(Interface, 0x38);                      // --- CLOCK
  }
  
  Interface.InputStatus = IStatus;
  return(IStatus);
}

// GetAnalog ----------------------------------------------------------------
// 
// Abfrage eines der beiden Analogeingnge
//
// ----------------------------------------------------------------------------

DWORD GetAnalog(ftiDCB &Interface, DWORD AnalogNr)
{
  if(Interface.hCom == 0) return(ftiFehler);
	if(Interface.PortID == 0)
    if(Interface.ECount == 8) return(GetAnalogCOM1(Interface, AnalogNr)); 
		else return(GetAnalogCOM2(Interface, AnalogNr));
  else
    if(Interface.PortID == 1) return(GetAnalogRT(Interface, AnalogNr));
      else return(GetAnalogLPT(Interface, AnalogNr));
}


// GetAnalogCOM1 ----------------------------------------------------------------
//
// Abfrage eines der beiden COM-Analogeingnge
//
// -----------------------------------------------------------------------------

DWORD GetAnalogCOM1(ftiDCB &Interface, DWORD AnalogNr)
{
  HANDLE hCom = Interface.hCom;
  unsigned char Befehl[2];
  unsigned char IAStatus[3];
  DWORD res, AnalogWert;
  DWORD Geschrieben, Gelesen;


  if(AnalogNr == 0) Befehl[0] = '\xC5'; else Befehl[0] = '\xC9';
  Befehl[1] = unsigned char(Interface.OutputStatus & 0xFF);

  res = WriteFile(hCom, Befehl, 2, &Geschrieben, NULL);
  res = ReadFile(hCom, IAStatus, 3, &Gelesen, NULL);
  if(Gelesen == 0) {
		 Interface.InputStatus  = ftiFehler; 
	                  return(ftiFehler);	}

  AnalogWert = (unsigned char(IAStatus[1])<<8) + unsigned char(IAStatus[2]);
  if (AnalogWert < 0 || AnalogWert > 2048) AnalogWert = 2048;
	Interface.InputStatus = IAStatus[0];
  return (AnalogWert);
}


// GetAnalogCOM2 ----------------------------------------------------------------
//
// Abfrage eines der beiden COM-Analogeingnge
//
// -----------------------------------------------------------------------------

DWORD GetAnalogCOM2(ftiDCB &Interface, DWORD AnalogNr)
{
  HANDLE hCom = Interface.hCom;
  unsigned char Befehl[3];
  unsigned char IAStatus[4];
  DWORD res, AnalogWert;
  DWORD Geschrieben, Gelesen;


  if(AnalogNr == 0) Befehl[0] = '\xC6'; else Befehl[0] = '\xCA';
  Befehl[1] = unsigned char(Interface.OutputStatus & 0xFF);
  Befehl[2] = unsigned char((Interface.OutputStatus & 0xFF00)>>8);

  res = WriteFile(hCom, Befehl, 3, &Geschrieben, NULL);
  res = ReadFile(hCom, IAStatus, 4, &Gelesen, NULL);
  if(Gelesen == 0) {
		 Interface.InputStatus  = ftiFehler; 
	                  return(ftiFehler);	}

  AnalogWert = (unsigned char(IAStatus[2])<<8) + unsigned(IAStatus[3]);
  if (AnalogWert < 0 || AnalogWert > 2048) AnalogWert = 2048;
	Interface.InputStatus = unsigned char(IAStatus[0]) + (unsigned char(IAStatus[1])<<8);;
  return (AnalogWert);
}


// GetAnalogLPT ----------------------------------------------------------------
//
// Abfrage eines der beiden LPT-Analogeingnge
//
// -----------------------------------------------------------------------------

DWORD GetAnalogLPT(ftiDCB &Interface, DWORD AnalogNr)
{
  DWORD i, AWert = 0;

// --- Trigger X bzw. Y ----
  if(AnalogNr == 0) SetLPT(Interface, 0xA0); else SetLPT(Interface, 0x90);
  SetLPT(Interface, 0x38);              // --- CLOCK

// --- Zhlschleife ----

  i = 0;
  do {                                  // --- Loop bei Input = 0xF8
    i++;
    if(i == Interface.LPTAnalog) {
      AWert += 1;
      i = 0;
      if(AWert >= 2048) break;
    }
  }
  while (GetLPT(Interface) & 0x80);     // --- Ende bei Input = 0x80
  
  return(AWert);
}



// umGetInput -------------------------------------------------------------------
// 
// Lesen Status Eingang InputNr
//
// -----------------------------------------------------------------------------

BOOL __stdcall umGetInput(ftiDCB &Interface, DWORD InputNr)
{
  DWORD res;

  res = ftiFehler;
	if (Interface.hCom == 0) return(res);
	
	res = Interface.InputStatus;
  return((res & EMaske[InputNr]) ? -1 : 0);
}

// umSetMotor --------------------------------------------------------------------
//
// Schalten des (Motor)Ausganges MotorNr
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetMotor(ftiDCB &Interface, DWORD MotorNr, DWORD Direction)
{
  DWORD MotorStatus, res;

	res = ftiFehler;
	if (Interface.hCom == 0) return(res);

  MotorStatus = Interface.OutputStatus & MAus[MotorNr];
  switch (Direction) {
  case ftiAus: break;
  case ftiLinks:  MotorStatus |= MLinks[MotorNr]; break;
  case ftiRechts: MotorStatus |= MRechts[MotorNr]; break;
  default: MotorStatus = 0;
  }
  Interface.OutputStatus = MotorStatus;
	res = Interface.InputStatus;
  return(res);
}


// umSetLamp --------------------------------------------------------------------
//
// Schalten des LampenAusganges LampNr 1 - 8 (16)
//
// -----------------------------------------------------------------------------

DWORD __stdcall umSetLamp(ftiDCB &Interface, DWORD LampNr, DWORD OnOff)
{
	DWORD res;
	res = ftiFehler;
	if (Interface.hCom == 0) return(res);

  if(Interface.PortID)       // --- paralleles Interface, PortID != 0
	{
		Interface.OutputStatus &= ~EMaske[LampNr];   // --- Lschen LampNr
		if(OnOff) Interface.OutputStatus |= EMaske[LampNr];
	}
	else                              // --- serielles Interface (PortID = 0)
	{
		Interface.OutputStatus |= EMaske[LampNr];
		if(OnOff) Interface.OutputStatus &= ~EMaske[LampNr];
	}
	res = Interface.InputStatus;
  return(res);
}



// GetLPT ----------------------------------------------------------------------
//
// Lesen eines WORD (2 Byte : Integer) vom aktuellen LPT-Port+1
//
// -----------------------------------------------------------------------------

DWORD GetLPT(ftiDCB &Interface)
{
  WORD res;
  WORD portid = WORD(Interface.PortID) + 1;

    __asm {
      push dx
      mov dx, portid
      in ax, dx
      mov res, ax
      pop dx
    }

  return(res);
}

// SetLPT       ----------------------------------------------------------------
//
// Schreiben eines WORD (2 Byte : Integer) auf den aktuellen LPT-Port
//
// -----------------------------------------------------------------------------

VOID  SetLPT(ftiDCB &Interface, WORD Wert)
{
  WORD portid = WORD(Interface.PortID);

  for (DWORD i=1; i<=Interface.LPTDelay; i++) 
  {
    __asm {
      push dx
      mov dx, portid
      mov ax, Wert
      out dx, ax
      pop dx
    } 
  }

}



// DllMain ---------------------------------------------------------------------------------
//
// Hauptschleife bei Start und Ende
//
// -----------------------------------------------------------------------------------------

BOOL WINAPI DllMain (HINSTANCE hDLL, DWORD dwReason, LPVOID)
{
	switch(dwReason) {
		case DLL_PROCESS_DETACH:
			umCloseInterface(ftiSave);
			break;
	}
	return TRUE;
}